import http from 'http';
import https from 'https';
import {constants} from 'http2';
import * as v from 'valibot';
import {
	AllowedMiddlewareSpecification,
	BackendState,
	Middleware,
	RequestContext, RequestDecorator,
	Response,
} from '../../common';
import {CanPatchSpec, DELTA_SCHEMA, PATCH_CONTENT_MAP_TYPE, PatchContentType, Resource} from '../../../common';
import {
	handleGetRoot, handleOptions,
} from './handlers/default';
import {
	handleCreateItem,
	handleDeleteItem,
	handleEmplaceItem,
	handleGetCollection,
	handleGetItem,
	handlePatchItem,
} from './handlers/resource';
import {getBody, isTextMediaType} from './utils';
import {decorateRequestWithBackend} from './decorators/backend';
import {decorateRequestWithMethod} from './decorators/method';
import {decorateRequestWithUrl} from './decorators/url';
import {ErrorPlainResponse, PlainResponse} from './response';
import EventEmitter from 'events';
import {DataSource} from '../../data-source';

type RequiredResource = Required<Pick<RequestContext, 'resource'>>['resource'];

interface ResourceWithDataSource extends Omit<RequiredResource, 'dataSource'> {
	dataSource: DataSource;
}

interface ResourceRequestContext extends Omit<RequestContext, 'resource'> {
	resource: ResourceWithDataSource;
}

declare module '../../common' {
	interface RequestContext extends http.IncomingMessage {
		body?: unknown;
	}

	interface Middleware<Req extends ResourceRequestContext = ResourceRequestContext, Res extends NodeJS.EventEmitter = NodeJS.EventEmitter> {
		(req: Req, res: Res): undefined | Response | Promise<undefined | Response>;
	}
}

const constructPostSchema = <T extends v.BaseSchema>(resource: Resource<T>) => {
	return resource.schema;
};

const constructPutSchema = <T extends v.BaseSchema>(resource: Resource<T>, mainResourceId?: string) => {
	if (typeof mainResourceId === 'undefined') {
		return resource.schema;
	}

	const schema = resource.schema.type === 'object' ? resource.schema as unknown as v.ObjectSchema<any> : resource.schema;
	const idAttr = resource.state.shared.get('idAttr') as string;
	const idConfig = resource.state.shared.get('idConfig') as any;
	return (
		schema.type === 'object'
			? v.merge([
				schema as v.ObjectSchema<any>,
				v.object({
					[idAttr]: v.transform(
						v.any(),
						input => idConfig!.serialize(input),
						v.literal(mainResourceId)
					)
				})
			])
			: schema
	);
};

const constructPatchSchema = <T extends v.BaseSchema>(resource: Resource<T>) => {
	const schema = resource.schema.type === 'object' ? resource.schema as unknown as v.ObjectSchema<any> : resource.schema;

	if (resource.schema.type !== 'object') {
		return resource.schema;
	}

	const schemaChoices = {
		merge: v.partial(
			schema as v.ObjectSchema<any>,
			(schema as v.ObjectSchema<any>).rest,
			(schema as v.ObjectSchema<any>).pipe
		),
		delta: v.array(DELTA_SCHEMA),
	}

	const selectedSchemaChoices = Object.entries(schemaChoices)
		.filter(([key]) => resource.state.canPatch[key as CanPatchSpec])
		.map(([, value]) => value);

	return v.union(selectedSchemaChoices);
};
// TODO add a way to define custom middlewares
const defaultCollectionMiddlewares: AllowedMiddlewareSpecification[] = [
	{
		method: 'GET',
		middleware: handleGetCollection,
		allowed: (resource) => resource.state.canFetchCollection,
	},
	{
		method: 'POST',
		middleware: handleCreateItem,
		allowed: (resource) => resource.state.canCreate,
		constructBodySchema: constructPostSchema,
	},
];

const defaultItemMiddlewares: AllowedMiddlewareSpecification[] = [
	{
		method: 'GET',
		middleware: handleGetItem,
		allowed: (resource) => resource.state.canFetchItem,
	},
	{
		method: 'PUT',
		middleware: handleEmplaceItem,
		constructBodySchema: constructPutSchema,
		allowed: (resource) => resource.state.canEmplace,
	},
	{
		method: 'PATCH',
		middleware: handlePatchItem,
		constructBodySchema: constructPatchSchema,
		allowed: (resource) => resource.state.canPatch.merge || resource.state.canPatch.delta,
	},
	{
		method: 'DELETE',
		middleware: handleDeleteItem,
		allowed: (resource) => resource.state.canDelete,
	},
];

export interface CreateServerParams {
	basePath?: string;
	host?: string;
	cert?: string;
	key?: string;
	requestTimeout?: number;
	// CQRS
	streamResponses?: boolean;
}

class CqrsEventEmitter extends EventEmitter {

}

interface ServerState {
	requestDecorators: Set<RequestDecorator>;
}

export interface Server {
	readonly listening: boolean;
	on(event: string, cb: (...args: unknown[]) => unknown): this;
	close(callback?: (err?: Error) => void): this;
	listen(...args: Parameters<http.Server['listen']>): this;
	requestDecorator(requestDecorator: RequestDecorator): this;
	defaultErrorHandler(): this;
}

export const createServer = (backendState: BackendState, serverParams = {} as CreateServerParams) => {
	const state: ServerState = {
		requestDecorators: new Set<RequestDecorator>(),
	};

	const isHttps = 'key' in serverParams && 'cert' in serverParams;
	const theRes = new CqrsEventEmitter();

	const server = isHttps
		? https.createServer({
			key: serverParams.key,
			cert: serverParams.cert,
			requestTimeout: serverParams.requestTimeout,
		})
		: http.createServer({
			requestTimeout: serverParams.requestTimeout,
		});

	const handleMiddlewares = async (currentHandlerState: Awaited<ReturnType<Middleware>>, currentMiddleware: AllowedMiddlewareSpecification, req: ResourceRequestContext) => {
		const { method: middlewareMethod, middleware, constructBodySchema} = currentMiddleware;
		const effectiveMethod = req.method === 'HEAD' ? 'GET' : req.method;

		if (effectiveMethod !== middlewareMethod) {
			return currentHandlerState;
		}

		if (typeof currentHandlerState !== 'undefined') {
			return currentHandlerState;
		}

		if (typeof constructBodySchema === 'function') {
			const bodySchema = constructBodySchema(req.resource, req.resourceId);
			const contentTypeHeader = req.headers['content-type'] ?? 'application/octet-stream';
			const fragments = contentTypeHeader.replace(/\s+/g, ' ').split(';');
			const mediaType = fragments[0];
			const charsetParam = (
				fragments
					.map((s) => s.trim())
					.find((f) => f.startsWith('charset='))

				?? (
					isTextMediaType(mediaType)
						? 'charset=utf-8'
						: 'charset=binary'
				)
			);
			const [_charsetKey, charsetRaw] = charsetParam.split('=').map((s) => s.trim());
			const charset = (
				(
					(charsetRaw.startsWith('"') && charsetRaw.endsWith('"'))
					|| (charsetRaw.startsWith("'") && charsetRaw.endsWith("'"))
				)
					? charsetRaw.slice(1, -1).trim()
					: charsetRaw.trim()
			) ?? (isTextMediaType(mediaType) ? 'utf-8' : 'binary');

			if (effectiveMethod === 'PATCH') {
				const validPatchTypes = Object.entries(req.resource.state.canPatch)
					.filter(([, allowed]) => allowed)
					.map(([patchType]) => patchType);

				const validPatchContentTypes = Object.entries(PATCH_CONTENT_MAP_TYPE)
					.filter(([ patchType]) => validPatchTypes.includes(patchType))
					.map(([contentType ]) => contentType);

				const isPatchEnabled = req.resource.state.canPatch[PATCH_CONTENT_MAP_TYPE[mediaType as PatchContentType]];
				if (!isPatchEnabled) {
					throw new ErrorPlainResponse('invalidResourcePatchType', {
						statusCode: constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE,
						res: theRes,
						headers: {
							'Accept-Patch': validPatchContentTypes.join(', '),
						},
					});
				}
			}

			const theBodyBuffer = await getBody(req);
			const encodingPair = req.backend.app.charsets.get(charset);
			if (typeof encodingPair === 'undefined') {
				throw new ErrorPlainResponse('unableToDecodeResource', {
					statusCode: constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE,
					res: theRes,
				});
			}
			const deserializerPair = req.backend.app.mediaTypes.get(mediaType);
			if (typeof deserializerPair === 'undefined') {
				throw new ErrorPlainResponse('unableToDeserializeResource', {
					statusCode: constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE,
					res: theRes,
				});
			}
			const theBodyStr = encodingPair.decode(theBodyBuffer);
			const theBody = deserializerPair.deserialize(theBodyStr);
			try {
				req.body = await v.parseAsync(bodySchema, theBody, {abortEarly: false, abortPipeEarly: false});
			} catch (errRaw) {
				const err = errRaw as v.ValiError;
				// todo use error message key for each method
				// TODO better error reporting, localizable messages
				// TODO handle error handlers' errors
				if (Array.isArray(err.issues)) {
					if (req.method === 'PATCH' && req.headers['content-type']?.startsWith('application/json-patch+json')) {
						throw new ErrorPlainResponse('invalidResourcePatch', {
							statusCode: constants.HTTP_STATUS_BAD_REQUEST,
							body: err.issues.map((i) => (
								`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
							)),
							res: theRes,
						});
					}

					throw new ErrorPlainResponse('invalidResource', {
						statusCode: constants.HTTP_STATUS_BAD_REQUEST,
						body: err.issues.map((i) => (
							`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
						)),
						res: theRes,
					});
				}
			}
		}

		const result = await middleware(req, theRes);

		// HEAD is just GET without the response body
		if (req.method === 'HEAD' && result instanceof PlainResponse) {
			const { body: _, ...etcResult } = result;

			return new PlainResponse({
				...etcResult,
				res: theRes,
			});
		}

		return result;
	};

	const processRequest = (middlewares: AllowedMiddlewareSpecification[]) => async (req: ResourceRequestContext) => {
		if (req.url === '/' || req.url === '') {
			return handleGetRoot(req, theRes);
		}

		const { resource } = req;
		if (typeof resource === 'undefined') {
			throw new ErrorPlainResponse('resourceNotFound', {
				statusCode: constants.HTTP_STATUS_NOT_FOUND,
				res: theRes,
			});
		}

		if (req.method === 'OPTIONS') {
			return handleOptions(middlewares)(req, theRes);
		}

		if (typeof resource.dataSource === 'undefined') {
			throw new ErrorPlainResponse('unableToBindResourceDataSource', {
				statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
				res: theRes,
			});
		}

		try {
			await resource.dataSource.initialize();
		} catch (cause) {
			throw new ErrorPlainResponse(
				'unableToInitializeResourceDataSource',
				{
					cause,
					statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
					res: theRes,
				}
			);
		}

		const middlewareResponse = await middlewares.reduce<ReturnType<Middleware>>(
			async (currentHandlerStatePromise, currentMiddleware) => {
				const currentHandlerState = await currentHandlerStatePromise;
				return await handleMiddlewares(currentHandlerState, currentMiddleware, req);
			},
			Promise.resolve<ReturnType<Middleware>>(undefined)
		) as Awaited<ReturnType<Middleware>>;

		if (typeof middlewareResponse === 'undefined') {
			throw new ErrorPlainResponse('resourceNotFound', {
				statusCode: constants.HTTP_STATUS_NOT_FOUND,
				res: theRes,
			});
		}

		return middlewareResponse as Awaited<ReturnType<Middleware>>
	};

	const defaultRequestDecorators = [
		decorateRequestWithMethod,
		decorateRequestWithUrl(serverParams),
		decorateRequestWithBackend(backendState),
	];

	const decorateRequest = async (reqRaw: http.IncomingMessage) => {
		const effectiveRequestDecorators = [
			...defaultRequestDecorators,
			...Array.from(state.requestDecorators),
		];

		return await effectiveRequestDecorators.reduce(
			async (resultRequestPromise, decorator) => {
				const resultRequest = await resultRequestPromise;
				const decoratedRequest = await decorator(resultRequest);
				// TODO log decorators
				return decoratedRequest;
			},
			Promise.resolve(reqRaw as RequestContext)
		);
	};

	const handleMiddlewareError = (processRequestErrRaw: Error) => (resourceReq: ResourceRequestContext, res: http.ServerResponse<RequestContext>) => {
		const finalErr = processRequestErrRaw as ErrorPlainResponse;
		const headers = finalErr.headers ?? {};
		const language = resourceReq.cn.language ?? resourceReq.backend.cn.language;
		const mediaType = resourceReq.cn.mediaType ?? resourceReq.backend.cn.mediaType;
		const charset = resourceReq.cn.charset ?? resourceReq.backend.cn.charset;
		let encoded: Buffer | undefined;
		let serialized;
		try {
			serialized = typeof finalErr.body !== 'undefined' ? mediaType.serialize(finalErr.body) : undefined;
		} catch (cause) {
			res.statusMessage = language.statusMessages['unableToSerializeResponse']?.replace(
				/\$RESOURCE/g,
				resourceReq.resource.state.itemName) ?? '';
			res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR);
			res.end();
			return;
		}

		try {
			encoded = typeof serialized !== 'undefined' ? charset.encode(serialized) : undefined;
		} catch (cause) {
			res.statusMessage = language.statusMessages['unableToEncodeResponse']?.replace(/\$RESOURCE/g,
				resourceReq.resource.state.itemName) ?? '';
			res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR);
			res.end();
			return;
		}

		headers['Content-Type'] = [
			mediaType.name,
			typeof serialized !== 'undefined' ? `charset=${charset.name}` : '',
		]
			.filter((s) => s.length > 0)
			.join('; ');

		res.statusMessage = language.statusMessages[
			finalErr.statusMessage ?? 'internalServerError'
		]?.replace(/\$RESOURCE/g,
			resourceReq.resource.state.itemName);
		res.writeHead(finalErr.statusCode ?? constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, headers);
		if (typeof encoded !== 'undefined') {
			res.end(encoded);
			return;
		}
		res.end();
	};

	const handleRequest = async (reqRaw: RequestContext, res: http.ServerResponse<RequestContext>) => {
		const plainReq = await decorateRequest(reqRaw); // TODO add type safety here, put handleGetRoot as its own middleware as it does not concern over any resource
		if (typeof plainReq.resource !== 'undefined') {
			const resourceReq = plainReq as ResourceRequestContext;
			const language = resourceReq.cn.language ?? resourceReq.backend.cn.language;
			const mediaType = resourceReq.cn.mediaType ?? resourceReq.backend.cn.mediaType;
			const charset = resourceReq.cn.charset ?? resourceReq.backend.cn.charset;

			// TODO custom middlewares
			const effectiveMiddlewares = (
				typeof resourceReq.resourceId === 'string'
					? defaultItemMiddlewares
					: defaultCollectionMiddlewares
			);
			const middlewares = effectiveMiddlewares.filter((m) => m.allowed(resourceReq.resource));
			// TODO listen to res.on('response')
			const processRequestFn = processRequest(middlewares);
			let middlewareState: Response;
			try {
				middlewareState = await processRequestFn(resourceReq) as any; // TODO fix this
			} catch (processRequestErrRaw) {
				handleMiddlewareError(processRequestErrRaw as Error)(resourceReq, res);
				return;
			}

			//  TODO extract to separate function
			const headers: Record<string, string> = {
				...(
					middlewareState.headers ?? {}
				),
				'Content-Language': language.name,
			};
			if (middlewareState instanceof http.ServerResponse) {
				// TODO streaming responses
				middlewareState.writeHead(constants.HTTP_STATUS_ACCEPTED, headers);
				return;
			}
			if (middlewareState instanceof PlainResponse) {
				let encoded: Buffer | undefined;
				if (typeof middlewareState.body !== 'undefined') {
					let serialized;
					try {
						serialized = mediaType.serialize(middlewareState.body);
					} catch (cause) {
						const headers: Record<string, string> = {
							'Content-Language': language.name,
						};
						if (resourceReq.method === 'POST') {
							headers['Accept-Post'] = Array.from(resourceReq.backend.app.mediaTypes.keys()).join(',');
						} else if (resourceReq.method === 'PATCH') {
							headers['Accept-Patch'] = Array.from(resourceReq.backend.app.mediaTypes.keys()).map((m) => {
								const [mimeType, mimeSubtype] = m.split('/');

								// TODO accept only patch document type from request
								// TODO implement Vary header (which headers influenced the request)
								// TODO implement OPTIONS method for determining the accepted media types and languages
								// TODO configure strict and lax accept behavior for content negotiation
								return `${mimeType}/merge-patch+${mimeSubtype}`;
							}).join(',');
						}
						res.statusMessage = language.statusMessages['unableToSerializeResponse']?.replace(
							/\$RESOURCE/g,
							resourceReq.resource!.state.itemName) ?? '';
						res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, headers);
						res.end();
						return;
					}

					try {
						encoded = charset.encode(serialized);
					} catch (cause) {
						res.statusMessage = language.statusMessages['unableToEncodeResponse']?.replace(/\$RESOURCE/g,
							resourceReq.resource!.state.itemName) ?? '';
						res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
							'Content-Language': language.name,
						});
						res.end();
						return;
					}

					headers['Content-Type'] = [
						mediaType.name,
						`charset=${charset.name}`,
					].join('; ');
				}

				const statusMessageKey = middlewareState.statusMessage ? language.statusMessages[middlewareState.statusMessage] : undefined;
				res.statusMessage = statusMessageKey?.replace(/\$RESOURCE/g, resourceReq.resource!.state.itemName) ?? '';
				res.writeHead(middlewareState.statusCode, headers);
				if (typeof encoded !== 'undefined') {
					res.end(encoded);
					return;
				}
				res.end();
				return;
			}

			if (middlewares.length > 0) {
				res.statusMessage = resourceReq.backend.cn.language.statusMessages.methodNotAllowed.replace(/\$RESOURCE/g,
					resourceReq.resource!.state.itemName) ?? '';
				res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
					Allow: middlewares.map((m) => m.method).join(', '),
					'Content-Language': resourceReq.backend.cn.language.name,
				});
				res.end();
				return;
			}
			res.statusMessage = resourceReq.backend.cn.language.statusMessages.urlNotFound.replace(/\$RESOURCE/g,
				resourceReq.resource!.state.itemName) ?? '';
			res.writeHead(constants.HTTP_STATUS_NOT_FOUND, {
				'Content-Language': resourceReq.backend.cn.language.name,
			});
			res.end();
			return;
		}

		res.statusMessage = reqRaw.backend.cn.language.statusMessages.notImplemented.replace(/\$RESOURCE/g,
			reqRaw.resource!.state.itemName) ?? '';
		res.writeHead(constants.HTTP_STATUS_NOT_IMPLEMENTED, {
			'Content-Language': reqRaw.backend.cn.language.name,
		});
		res.end();
	};

	server.on('request', handleRequest);

	return {
		get listening() { return server.listening },
		listen(...args: Parameters<Server['listen']>) {
			server.listen(...args);
			return this;
		},
		close(callback?: (err?: Error) => void) {
			server.close(callback);
			return this;
		},
		on(...args: Parameters<Server['on']>) {
			server.on(args[0], args[1]);
			return this;
		},
		requestDecorator(requestDecorator: RequestDecorator) {
			state.requestDecorators.add(requestDecorator);
			return this;
		},
		defaultErrorHandler() {
			return this;
		}
	} satisfies Server;

	// return server;
}
