import {constants} from 'http2';
import {AllowedMiddlewareSpecification, Middleware} from '../../../common';
import {LinkMap} from '../utils';
import {PlainResponse, ErrorPlainResponse} from '../response';
import {PATCH_CONTENT_MAP_TYPE} from '../../../../common';

export const handleGetRoot: Middleware = (req, res) => {
	const { backend, basePath } = req;

	const data = {
		name: backend.app.name
	};

	const registeredResources = Array.from(backend.app.resources);
	const availableResources = registeredResources.filter((r) => (
		r.state.canFetchCollection
		|| r.state.canCreate
	));

	const headers: Record<string, string> = {};
	if (availableResources.length > 0) {
		// https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Link
		headers['Link'] = new LinkMap(
			availableResources.map((r) => ({
				url: `${basePath}/${r.state.routeName}`,
				params: {
					rel: 'related',
					name: r.state.routeName,
				},
			}))
		)
			.toString();
	}

	return new PlainResponse({
		headers,
		statusMessage: 'ok',
		statusCode: constants.HTTP_STATUS_OK,
		body: data,
		res,
	});
};

export const handleOptions = (middlewares: AllowedMiddlewareSpecification[]): Middleware => (req, res) => {
	if (middlewares.length > 0) {
		const allowedMethods = middlewares.flatMap((m) => m.method === 'GET' ? [m.method, 'HEAD'] : [m.method]);
		const headers: Record<string, string> = {
			'Allow': allowedMethods.join(', '),
		};
		if (allowedMethods.includes('PATCH')) {
			const validPatchTypes = Object.entries(req.resource.state.canPatch)
				.filter(([, allowed]) => allowed)
				.map(([patchType]) => patchType);

			const validPatchContentTypes = Object.entries(PATCH_CONTENT_MAP_TYPE)
				.filter(([, patchType]) => validPatchTypes.includes(patchType))
				.map(([contentType ]) => contentType);

			if (validPatchContentTypes.length > 0) {
				headers['Accept-Patch'] = validPatchContentTypes.join(', ');
			}
		}
		return new PlainResponse({
			headers,
			statusMessage: 'provideOptions',
			statusCode: constants.HTTP_STATUS_NO_CONTENT,
			res,
		});
	}

	// TODO add option for custom error handler
	throw new ErrorPlainResponse('methodNotAllowed', {
		statusCode: constants.HTTP_STATUS_METHOD_NOT_ALLOWED,
		res,
	});
};
