import * as v from 'valibot';
import {PatchContentType} from './media-type';
import {DataSource, ResourceIdConfig} from '../backend';

export const CAN_PATCH_VALID_VALUES = ['merge', 'delta'] as const;

export type CanPatchSpec = typeof CAN_PATCH_VALID_VALUES[number];

export const PATCH_CONTENT_MAP_TYPE: Record<PatchContentType, CanPatchSpec> = {
	'application/merge-patch+json': 'merge',
	'application/json-patch+json': 'delta',
};

type CanPatchObject = Record<CanPatchSpec, boolean>;

export interface ResourceState<
	ItemName extends string = string,
	RouteName extends string = string
> {
	shared: Map<string, unknown>;
	relationships: Set<Resource>;
	itemName: ItemName;
	routeName: RouteName;
	canCreate: boolean;
	canFetchCollection: boolean;
	canFetchItem: boolean;
	canPatch: CanPatchObject;
	canEmplace: boolean;
	canDelete: boolean;
}

type CanPatch = boolean | Partial<CanPatchObject> | CanPatchSpec[];

export interface Resource<
	Schema extends v.BaseSchema = v.BaseSchema,
	CurrentName extends string = string,
	CurrentRouteName extends string = string,
	CurrentIdAttr extends string = string,
	IdSchema extends v.BaseSchema = v.BaseSchema
> {
	schema: Schema;
	state: ResourceState<CurrentName, CurrentRouteName>;
	name<NewName extends CurrentName>(n: NewName): Resource<Schema, NewName, CurrentRouteName, CurrentIdAttr, IdSchema>;
	route<NewRouteName extends CurrentRouteName>(n: NewRouteName): Resource<Schema, CurrentName, NewRouteName, CurrentIdAttr, IdSchema>;
	canFetchCollection(b?: boolean): this;
	canFetchItem(b?: boolean): this;
	canCreate(b?: boolean): this;
	canPatch(b?: CanPatch): this;
	canEmplace(b?: boolean): this;
	canDelete(b?: boolean): this;
	relatedTo<RelatedSchema extends v.BaseSchema>(resource: Resource<RelatedSchema>): this;
	dataSource?: DataSource;
	id<NewIdAttr extends CurrentIdAttr, TheIdSchema extends IdSchema>(
		newIdAttr: NewIdAttr,
		params: ResourceIdConfig<TheIdSchema>
	): Resource<Schema, CurrentName, CurrentRouteName, NewIdAttr, TheIdSchema>;
}

export const resource = <
	Schema extends v.BaseSchema,
	CurrentName extends string = string,
	CurrentRouteName extends string = string
>(schema: Schema): Resource<Schema, CurrentName, CurrentRouteName> => {
	const resourceState = {
		shared: new Map(),
		relationships: new Set<Resource>(),
		canCreate: false,
		canFetchCollection: false,
		canFetchItem: false,
		canPatch: {
			merge: false,
			delta: false,
		},
		canEmplace: false,
		canDelete: false,
	} as ResourceState<CurrentName, CurrentRouteName>;

	return {
		get state(): ResourceState<CurrentName, CurrentRouteName> {
			return Object.freeze({
				...resourceState
			}) as unknown as ResourceState<CurrentName, CurrentRouteName>;
		},
		canFetchCollection(b = true) {
			resourceState.canFetchCollection = b;
			return this;
		},
		canFetchItem(b = true) {
			resourceState.canFetchItem = b;
			return this;
		},
		canCreate(b = true) {
			resourceState.canCreate = b;
			return this;
		},
		canPatch(b = true as CanPatch) {
			if (typeof b === 'boolean') {
				resourceState.canPatch.merge = b;
				resourceState.canPatch.delta = b;
				return this;
			}

			if (typeof b === 'object') {
				if (Array.isArray(b)) {
					CAN_PATCH_VALID_VALUES.forEach((p) => {
						resourceState.canPatch[p] = b.includes(p);
					});
					return this;
				}
				if (b !== null) {
					CAN_PATCH_VALID_VALUES.forEach((p) => {
						resourceState.canPatch[p] = b[p] ?? false;
					});
				}
			}

			return this;
		},
		canEmplace(b = true) {
			resourceState.canEmplace = b;
			return this;
		},
		canDelete(b = true) {
			resourceState.canDelete = b;
			return this;
		},
		id(idName, config) {
			resourceState.shared.set('idAttr', idName);
			resourceState.shared.set('idConfig', config);
			return this;
		},
		fullText(attrName: string) {
			const fullTextAttrs = (resourceState.shared.get('fullText') ?? new Set()) as Set<string>;
			fullTextAttrs.add(attrName);
			resourceState.shared.set('fullText', fullTextAttrs);
			return this;
		},
		name<NewName extends CurrentName>(n: NewName) {
			resourceState.itemName = n;
			return this as Resource<Schema, NewName, CurrentRouteName>;
		},
		route<NewRouteName extends CurrentRouteName>(n: NewRouteName) {
			resourceState.routeName = n;
			return this as Resource<Schema, CurrentName, NewRouteName>;
		},
		get itemName() {
			return resourceState.itemName;
		},
		get routeName() {
			return resourceState.routeName;
		},
		get schema() {
			return schema;
		},
		relatedTo<RelatedSchema extends v.BaseSchema>(resource: Resource<RelatedSchema>) {
			resourceState.relationships.add(resource);
			return this;
		},
	} as Resource<Schema, CurrentName, CurrentRouteName>;
};

export type ResourceType<R extends Resource> = v.Output<R['schema']>;
