import { readFile, writeFile } from 'fs/promises';
import { join } from 'path';
import { Resource, validation as v } from '@modal-sh/yasumi';
import { DataSource, ResourceIdConfig } from '@modal-sh/yasumi/backend';
import assert from 'assert';

export class ResourceNotPreparedError extends Error {}

export class ResourceIdNotDesignatedError extends Error {}

export class JsonLinesDataSource<
  Schema extends v.BaseSchema = v.BaseSchema,
  CurrentName extends string = string,
  CurrentRouteName extends string = string,
  Data extends object = v.Output<Schema>,
> implements DataSource<Data> {
  private path?: string;

  private resource?: Resource<Schema, CurrentName, CurrentRouteName>;

  data: Data[] = [];

  constructor(private readonly baseDir = '') {
    // noop
  }

  prepareResource(resource: Resource<Schema, CurrentName, CurrentRouteName>) {
    this.path = join(this.baseDir, `${resource.state.routeName}.jsonl`);
    resource.dataSource = resource.dataSource ?? this;
    const originalResourceId = resource.id;
    resource.id = <NewIdAttr extends string, NewIdSchema extends v.BaseSchema>(newIdAttr: NewIdAttr, params: ResourceIdConfig<NewIdSchema>) => {
      originalResourceId(newIdAttr, params);
      return resource as Resource<Schema, CurrentName, CurrentRouteName, NewIdAttr, NewIdSchema>;
    };
    this.resource = resource;
  }

  async initialize() {
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    try {
      const fileContents = await readFile(this.path, 'utf-8');
      const lines = fileContents.split('\n');
      this.data = lines.filter((l) => l.trim().length > 0).map((l) => JSON.parse(l));
    } catch (err) {
      await writeFile(this.path, '');
    }
  }

  async getTotalCount() {
    return this.data.length;
  }

  async getMultiple() {
    return [...this.data];
  }

  async newId() {
    const idConfig = this.resource?.state.shared.get('idConfig') as ResourceIdConfig<any>;
    assert(typeof idConfig !== 'undefined', new ResourceNotPreparedError());
    const theNewId = await idConfig.generationStrategy(this);
    return theNewId as string;
  }

  async getById(idSerialized: string) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    const id = idConfig.deserialize(idSerialized);
    const foundData = this.data.find((s) => (s as any)[idAttr] === id);

    if (foundData) {
      return {
        ...foundData
      };
    }

    return null;
  }

  async create(data: Data) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    const newData = {
      ...data
    } as Record<string, unknown>;

    if (idAttr in newData) {
      newData[idAttr] = idConfig.deserialize(newData[idAttr] as string);
    }

    const newCollection = [
      ...this.data,
      newData
    ];

    await writeFile(this.path, newCollection.map((d) => JSON.stringify(d)).join('\n'));

    return data as Data;
  }

  async delete(idSerialized: string) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    const oldDataLength = this.data.length;

    const id = idConfig.deserialize(idSerialized);
    const newData = this.data.filter((s) => !((s as any)[idAttr] === id));

    await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

    return oldDataLength !== newData.length;
  }

  async emplace(idSerialized: string, dataWithId: Data) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    const existing = await this.getById(idSerialized);
    const id = idConfig.deserialize(idSerialized);
    const { [idAttr]: idFromResource, ...data } = (dataWithId as any);
    const dataToEmplace = {
      [idAttr]: id,
      ...data,
    } as Data;

    if (existing) {
      const newData = this.data.map((d) => {
        if ((d as any)[idAttr] === id) {
          return dataToEmplace;
        }

        return d;
      });

      await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

      return [dataToEmplace, false] as [Data, boolean];
    }

    const newData = await this.create(dataToEmplace);
    return [newData, true] as [Data, boolean];
  }

  async patch(idSerialized: string, data: Partial<Data>) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    const existing = await this.getById(idSerialized);
    if (!existing) {
      return null;
    }

    const newItem = {
      ...existing,
      ...data,
    }

    const id = idConfig.deserialize(idSerialized);
    const newData = this.data.map((d) => {
      if ((d as any)[idAttr] === id) {
        return newItem;
      }

      return d;
    });

    await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));
    return newItem as Data;
  }
}
